# distutils: language = c++
# cython: language_level = 3
cimport numpy as np
import numpy as np

np.import_array()

from runtime.runtime cimport BatchMailBox, NullResult, Scheduler
from libc.stdio cimport printf
from libc.stdlib cimport malloc, free

## Types declaration
ctypedef int I_t
ctypedef double D_t

cdef lock Scheduler scheduler

cdef extern from *:
    """
    #include <algorithm>
    #include <iostream>
    template<class D, class I>
    class IndexComparator {
    private:
        const D *data;
        I split_dim, n_features;
    public:
        IndexComparator(const D *data, const I &split_dim, const I &n_features):
            data(data), split_dim(split_dim), n_features(n_features) {}
        bool operator()(const I &a, const I &b) const {
            D a_value = data[a * n_features + split_dim];
            D b_value = data[b * n_features + split_dim];
            return a_value == b_value ? a < b : a_value < b_value;
        }
    };
    template<class D, class I>
    void partition_node_indices(
        const D *data,
        I *node_indices,
        I start,
        I mid,
        I end,
        const I &split_dim,
        const I &n_features) {
        IndexComparator<D, I> index_comparator(data, split_dim, n_features);

        std::nth_element(
            node_indices + start,
            node_indices + mid,
            node_indices + end,
            index_comparator
        );

    }
    """
    void partition_node_indices[D, I](
                D *data_ptr,
                I *node_indices,
                I start,
                I mid,
                I end,
                I split_dim,
                I n_features
    ) nogil except +

cdef cypclass Node activable:
    """A KDTree Node"""

    D_t * point
    I_t n_dims
    active Node left
    active Node right

    __init__(self):
        self._active_result_class = NullResult
        self._active_queue_class = consume BatchMailBox(scheduler)
        self.left = NULL
        self.right = NULL

    void build_node(
            self,
            D_t * data_ptr,
            I_t * indices_ptr,
            I_t depth,
            I_t n_dims,
            I_t dim,
            I_t start,
            I_t end,
    ):
        cdef I_t i
        cdef I_t next_dim = (dim + 1) % n_dims
        cdef I_t mid = (start + end) // 2

        self.n_dims = n_dims

        if (depth < 0) or (end - start <= 1):
            return

        partition_node_indices(data_ptr, indices_ptr, start, mid, end, dim, n_dims)
        self.point = data_ptr + mid

        self.left = consume Node()
        self.right = consume Node()

        self.left.build_node(NULL,
                             data_ptr, indices_ptr,
                             depth - 1, n_dims, next_dim,
                             start, mid)
        self.right.build_node(NULL,
                              data_ptr, indices_ptr,
                              depth - 1, n_dims, next_dim,
                              mid, end)


cdef cypclass KDTree:
    """A KDTree based on asynchronous and parallel computations.
    
    It is similar to :class:`sklearn.neighbours.KDTree` but it is 
    using Cython+ features for asynchronous computations.
    
    Asynchronuous and parallel computations can be used when 
    constructing the nodes (see above) and when querying (WIP).
    
    This relies on a Cython+ runtime using actors.
    """

    I_t n # number of data
    I_t d # number of dimensions / features
    I_t depth # max_depth of the tree (to be unified with leaf_size)

    np.ndarray data_arr
    np.ndarray idx_array_arr

    active Node root

    D_t *data_ptr
    I_t *indices_ptr

    __init__(self,
             np.ndarray data,
             I_t depth,
             ):
        cdef I_t i
        cdef I_t n = data.shape[0]
        cdef I_t d = data.shape[1]

        self.n = n
        self.d = d
        self.depth = depth

        self.data_ptr = <D_t *> data.data
        self.indices_ptr = <I_t *> malloc(n * sizeof(I_t))
        for i in range(n):
            self.indices_ptr[i] = i

        self._recursive_build()

    void _recursive_build(self):
        # TODO: introducing a context manager for the runtime
        # would be nice here:
        #   ```
        #   with scheduler:
        #        self.root = ...
        #   ```
        global scheduler
        scheduler = Scheduler()

        self.root = consume Node()
        if self.root is NULL:
            printf("Error consuming node\n")

        # When object are activated (set as Actors), methods
        # are reified. When using those reified methods
        # a new argument is prepredend for the Promise,
        # which we aren't using using here, hence the extra NULL.
        self.root.build_node(NULL,
                             self.data_ptr,
                             self.indices_ptr,
                             depth, n_dims=d, dim=0, start=0, end=n)

        scheduler.finish()
        del scheduler

        for i in range(n):
            printf("X[indices[%d] = %d]= %f %f\n",
                   i, self.indices_ptr[i],
                   self.data_ptr[self.indices_ptr[i] * d],
                   self.data_ptr[self.indices_ptr[i] * d + 1])

    void __dealloc__(self):
        printf("Deallocating KDTree datastructures\n")
        free(self.data_ptr)
        free(self.indices_ptr)
        printf("Done deallocating KDTree datastructures\n")
        # XXX It sometimes blocks over here and then segfaults

cdef public int main() nogil:
    # Entry point for the compiled binary file
    printf("main: called\n")
    tree = KDTree()
    printf("main: done\n")
    return 0
